#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  This script provides a basic Tk Gui to do the logical processing
#  
#  See included OSCmessages.txt and filestructure.txt for more info
#  Copyright 2018 Zinc
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  
# ~ from osc4py3.as_allthreads import *
from osc4py3.as_allthreads import *
from osc4py3 import oscbuildparse
from osc4py3 import oscmethod as osm
from guizero import App, Text, PushButton, Combo, Slider, Box, Window, Picture
from time import sleep, time
from subprocess import Popen
import os.path, sys, threading

scriptPath = os.path.abspath(os.path.dirname(sys.argv[0]))
audioDir = scriptPath+"/audio/"
resourcesDir = scriptPath+"/static/" # this folder contains the images used on the UI
# next two dictionnaries are storing the path to images used on the UI to improve code readability
LED={"gray":resourcesDir+"led-rounded-h-grey.png", "green":resourcesDir+"led-rounded-h-green.png", "red":resourcesDir+"led-rounded-h-red.png", "yellow":resourcesDir+"led-rounded-h-yellow.png"}
ICONS={"next":resourcesDir+"rodentia-icons_media-skip-forward.png", "prev":resourcesDir+"rodentia-icons_media-skip-backward.png", "play":resourcesDir+"rodentia-icons_media-playback-start.png", "pause":resourcesDir+"rodentia-icons_media-playback-pause.png", "playSample":resourcesDir+"rodentia-icons_media-playback-start-green.png", "settings":resourcesDir+"preferences-system.png", "night":resourcesDir+"night.png", "day":resourcesDir+"day.png", "sensor":resourcesDir+"wireless-sensor.png", "sensorGray":resourcesDir+"wireless-sensor-gray.png", "pixel":resourcesDir+"pixel.png"}

startTime=time()
OSClisteningPort=9000 # must match the port used to send OSC by the wemosD1s
osc_udp_client("127.0.0.1", 9001, "puredata")# must match the IP and port used by puredata
osc_broadcast_client("10.0.0.255",8000, "broadcast")# must match broadcast IP defined in /etc/dnsmasq.conf and D1 listening port
myIp="10.0.0.1" # this Pi static IP is also defined in /etc/dnsmasq.conf
pdCommand = "pd -audioindev 3 -audiooutdev 3 -nogui "+scriptPath+"/pd/oloPd.pd&"
# pdPassthrough is used to pass OSC messages to pd without loading any files and thus play a synth for example
pdPassthrough={"synth":{"tracks":[{"title":"piste 1", "status":"normal","volume":0,"playable":True},{"title":"piste 2", "status":"normal","volume":0,"playable":True},{"title":"piste 3", "status":"normal","volume":0,"playable":True},{"title":"piste 4", "status":"normal","volume":0,"playable":True},{"title":"piste 5", "status":"normal","volume":0,"playable":True},{"title":"piste 6", "status":"normal","volume":0,"playable":True}]}}
pdThreeshold = 50 # under this threeshold, no data will be sent to the pd patch to reduce noise induced by ambiant light
updateUIdelay = 400 # time in ms between two UI  updates
offlineThreeshold = 3 # time in s after last communication to be declared offline
offlineTimer = 2000 # the sensor window will check for disconnected devices every offlineTimer ms
isPlaying = False
groups={} # will group every input of a sensor (and the associated tracks) to one button
# every track has a status which can be :
#       "normal" (default on load) : volume sent from sensors is mitigated by the sliders, then sent to puredata
#       "mute" : volume is fixed to 0, regardless of the sensor values
#       "manual" : volume is set entirely by this UI and will ignore the sensors
allTracksStatus="normal"
# expectedOSCmessagestell this script which OSC message to expect from sensors : [ {"address":"/OSCaddress","count": number_of_values_in_this_message) ]
# the script will then map them in order over the available tracks (ex : {"address":"/A", "count":2} will assign the first value within an /A message to track0 and the second to track1)
expectedOSCmessages=[{"address":"/A", "count":3}, {"address":"/B", "count":3}, {"address":"/C", "count":3}, {"address":"/D", "count":3}]
onlineDevices={}# will be populated by devices sending /myID to this Pi
# sampleTracks defines which tracks contains sounds to be played from the start when illuminated instead of being played continuously in loop mode
sampleTracks=(2,5,8,11) # will be overwritten by the sample tracks defined for the loaded song
defaultSampleTracks=(2,5,8,11) # if nothing is defined for the current song
sampleTracksDefinitions={"Reggae":(2,5,8,11),"DarkKnight":(2,5,8,11),"Hipitipop":(2,5,8,11),"HeYo":(2,5,8,11),"accordsCM":(0,1,2,3,4,5),"Choeurs":(0,1,2,3,4,5)}
sampleThreeshold = 400

# Setting up the GUI
app = App(layout="grid", bg="#4d4d4d", title="mixer", width=1000, height=610) # width will be overriden depending on track count of a given song
transportBox = Box(app, layout="grid", grid=[0,0,12,1]) # will contain play/pause, prev, next button and the song dropdown
statusText = Text(transportBox, color="white", grid=[0,2,3,1])# used for debug and to alert the user if an error should arise
buttons, groupButtons, sliders, labels, sampleTriggers, settingsButtons = [], [], [], [], [], [] # will contain the dynamically created widgets
defaultFaderValue = 70

# Read filestructure and populate the song list with every available track. See included fileStructure.txt for more info on this.
def updateSongList() :
    songList={}
    for folder in [d for d in os.listdir(audioDir) if os.path.isdir(os.path.join(audioDir, d))] :
        songList[folder]= {"tracks":[]} # name of the song will be the folder's name
        folderPath = audioDir + folder
        for wavFile in sorted([f for f in os.listdir(folderPath) if os.path.isfile(os.path.join(folderPath, f)) and os.path.join(folderPath, f).endswith(".wav")]): # alphabetical sort of the tracks
            title = wavFile.split("_")[1].split(".wav")[0] # get rid of anything before "_" and the extension
            songList[folder]["tracks"].append({"title":title, "file":audioDir+folder+"/"+wavFile, "status":"normal", "volume" : scaleVolume(1000,defaultFaderValue), "playable":True})
    songList.update(pdPassthrough) # add the synth output
    return songList

# Toggle a given track status between normal, mute and manual
def toggleTrack(trackNumber):
    status = songs[songSelection.value]["tracks"][trackNumber]["status"]
    if status == "normal" : setTrackStatus(trackNumber, "mute")
    elif status == "mute" :  setTrackStatus(trackNumber, "manual")
    elif status == "manual" : setTrackStatus(trackNumber, "normal")
    statusText.value = "piste "+songs[songSelection.value]["tracks"][trackNumber]["title"]+" basculée sur "+songs[songSelection.value]["tracks"][trackNumber]["status"]

# Alternate all tracks status between normal and manual. Used by keyboard shortcut.
def toggleAllTracks(event):
    global allTracksStatus
    if allTracksStatus == "normal":
        for trackNumber in range(0, len(songs[songSelection.value]["tracks"])) : setTrackStatus(trackNumber, "manual")
        allTracksStatus = "manual"
    elif allTracksStatus == "manual" :
        for trackNumber in range(0, len(songs[songSelection.value]["tracks"])) : setTrackStatus(trackNumber, "normal")
        allTracksStatus = "normal"

def toggleGroup(groupName) :
    if groups[groupName]["status"] == "normal":
        groups[groupName]["status"] = "mute"
        for track in groups[groupName]["tracks"]:setTrackStatus(track, "mute")
        for button in [b for b in groupButtons if b.text==groupName] : button.bg = "#ea4d5a"
        if groupName in onlineDevices : sendOSC(groupName,"/ledOFF")
    elif groups[groupName]["status"] == "mute":
        groups[groupName]["status"] = "normal"
        for track in groups[groupName]["tracks"]:setTrackStatus(track, "normal")
        for button in [b for b in groupButtons if b.text==groupName] : button.bg = "#4d4d4d"
        if groupName in onlineDevices : sendOSC(groupName,"/ledON")
    # ~ elif groups[groupName]["status"] == "manual":
        # ~ for track in groups[groupName]["tracks"]:setTrackStatus(track, "mute")
        # ~ groups[groupName]["status"] = "mute"
        # ~ for button in [b for b in groupButtons if b.text==groupName] : button.bg = "#f9f87d"
        
        

# Define what's happening when a track status is changed
def setTrackStatus(trackNumber, status) :
    global songs
    try : groupStatus = [groups[g]["status"] for g in groups if trackNumber in groups[g]["tracks"]][0]
    except IndexError: groupStatus = "normal" # this track doesn't belongs to any group, can happen if tracks and inputs counts doesn't align
    if status == "mute" :
        songs[songSelection.value]["tracks"][trackNumber]["status"] = "mute"
        buttons[trackNumber].image = LED["red"]
        sliders[trackNumber].value = 0
        sliders[trackNumber].enabled = False
        sendOSC("puredata","/volume", [trackNumber, 0])
    elif status == "manual" and groupStatus == "normal" :
        songs[songSelection.value]["tracks"][trackNumber]["status"] = "manual"
        buttons[trackNumber].image = LED["yellow"]
        sliders[trackNumber].enabled = True
        sliders[trackNumber].value = str(int(songs[songSelection.value]["tracks"][trackNumber]["volume"]/10))
    elif status == "normal" and groupStatus == "normal" :
        songs[songSelection.value]["tracks"][trackNumber]["status"] = "normal"
        ledColor = "green" if isPlaying else "gray"
        buttons[trackNumber].image = LED[ledColor]
        sliders[trackNumber].enabled = True
        sliders[trackNumber].value = str(int(songs[songSelection.value]["tracks"][trackNumber]["volume"]/10))

# called when releasing a fader on the UI
def changeVolume(eventData):
    trackNumber, volume = int(eventData.widget.name), int(eventData.widget.value)*10
    songs[songSelection.value]["tracks"][trackNumber]["volume"] = volume
    if songs[songSelection.value]["tracks"][trackNumber]["status"] != "mute" and trackNumber not in sampleTracks :
        sendOSC("puredata","/volume", [trackNumber, volume])
        return
    if songs[songSelection.value]["tracks"][trackNumber]["status"] != "mute" and trackNumber in sampleTracks : playSample(trackNumber, volume, True)

# Filter value below threeshold and either ask pd to play a sample or to adjust it's volume depending on if it's playing or not
def playSample(trackNumber, volume, force=False):
    if songs[songSelection.value]["tracks"][trackNumber]["status"] != "mute" :
        if force : sendOSC("puredata", "/playSample", [trackNumber, volume]) # called by UI, useful for testing sound configuration. Will play even on mute.
        if volume > sampleThreeshold and songs[songSelection.value]["tracks"][trackNumber]["playable"]: # if we passed the threeshold and the sample hasn't been played yet
            sendOSC("puredata", "/playSample", [trackNumber, volume]) # play it
            songs[songSelection.value]["tracks"][trackNumber]["playable"] = False # input need to pass under the threeshold before the sample can be played again
        elif volume < sampleThreeshold : songs[songSelection.value]["tracks"][trackNumber]["playable"] = True # like this
        if not songs[songSelection.value]["tracks"][trackNumber]["playable"]: sendOSC("puredata", "/volume", [trackNumber, volume]) # if sample is already playing, only adjust it's volume

# Define what to do when play/pause and skip next/prev commands are received
def transport(command):
    global isPlaying, allTracksStatus
    humanReadable={"playPause":"lecture/pause", "skipNext":"piste suivante", "skipPrev":"piste précédente"}
    # ~ statusText.value = "commande "+humanReadable[command]+" envoyée"
    if command == "playPause" : # toggle between playing mode and pause
        if isPlaying : 
            isPlaying = False
            playButton.image = ICONS["play"]
            for button in [b for b in buttons if b.image == LED["green"]] : button.image = LED["gray"]
            sendOSC("puredata","/pause",[])
        else :
            isPlaying = True
            playButton.image = ICONS["pause"]
            for button in [b for b in buttons if b.image == LED["gray"]] : button.image = LED["green"]
            sendOSC("puredata","/play",[])
    elif command == "skipNext" or command == "skipPrev" :
        songTitles = list(songs.keys())
        currentIndex = songTitles.index(songSelection.value)
        if command == "skipNext" : # wrap around indexes > len(songTitles)
            if currentIndex == len(songTitles)-1 : songSelection.value = songTitles[0]
            else : songSelection.value = songTitles[currentIndex+1]
        else : # wrap around indexes < 0
            if currentIndex == 0 : songSelection.value = songTitles[len(songTitles)-1]
            else : songSelection.value = songTitles[currentIndex-1]
        updateLayout(songSelection.value)
        allTracksStatus, isPlaying = "normal", False
        sendOSC("puredata","/pause",[])
        playButton.image = ICONS["play"]

# called by the night button on the main window
def calibrateAllSensors(event=None):
    sendOSC("broadcast", "/calibrateAmbiant")
    statusText.value = "calibration de la lumière ambiante effectuée"

# called by individual buttons on the sensor window
def calibrateSensor(OSCaddress, calibNight):
    OSCmessage = "/calibrateAmbiant" if calibNight else "/calibrateMax"
    sendOSC(OSCaddress,OSCmessage)
    UImessage = "de la nuit" if calibNight else "du jour"
    statusText.value = "calibration "+UImessage+" envoyée à "+OSCaddress

# Creates every buttons and sliders needed for the song currently loaded
def updateLayout(songTitle) :
    global buttons, groupButtons, sliders, labels, sampleTriggers, groups, sampleTracks
    for widget in buttons + groupButtons + sliders + labels + sampleTriggers : widget.destroy() # clear the old widgets
    buttons, groupButtons, sliders, labels, sampleTriggers, groups = [], [], [], [], [], {}
    # ~ if isPlaying : transport("playPause") # stop the playback
    if songTitle not in pdPassthrough : # if a song is loaded (not a synth)
        if songTitle in sampleTracksDefinitions : sampleTracks = sampleTracksDefinitions[songTitle]
        else : sampleTracks = defautSampleTracks
    for track in songs[songTitle]["tracks"] :
        trackNumber = songs[songTitle]["tracks"].index(track)
        labels.append(Text(app, text=track["title"], grid=[trackNumber,2], color="white")) # label the track by it's name
        buttons.append(PushButton(app, image=LED["gray"], text="", command=toggleTrack, args=[trackNumber], grid=[trackNumber,3])) # status button
        sliders.append(Slider(app,command=None, grid=[trackNumber,5], horizontal=False, start=100, end=0)); # fader will go vertically from 0 to 100
        sliders[len(sliders)-1].height = 300; sliders[len(sliders)-1].text_color = "white" # fader configuration (size, value, callback)
        sliders[len(sliders)-1].value = str(defaultFaderValue)
        sliders[len(sliders)-1].name=str(trackNumber) # hackish way to retrieve which slider was moved
        sliders[len(sliders)-1].when_left_button_released = changeVolume # hackish way to only use the callback when the slider is released
        if songTitle not in pdPassthrough :
            if trackNumber == 0 :  sendOSC("puredata","/sampler"); sleep(.05) # warn puredata that we are gonna load a sample
            if trackNumber in sampleTracks : sampleTriggers.append(PushButton(app, image=ICONS["playSample"], text="", align="top", command=playSample, args=[trackNumber, 1000, True], grid=[trackNumber,6]))
            sendOSC("puredata","/load", [trackNumber, songs[songTitle]["tracks"][trackNumber]["file"]])
        sleep(.05) # gives puredata the time to load a file
    # add a status button per group (1 group covers every inputs from one sensor)
    trackCount = len(songs[songTitle]["tracks"])
    horizontalOffset = 0
    [sendOSC(sensor, "/ledON") for sensor in onlineDevices.keys()] # turn on the talent's light of every sensor by default, since the mute groups are off
    for sensor in expectedOSCmessages : 
        buttonWidth = buttons[0].width * sensor["count"] # we will use the padding to size the button
        padx = buttonWidth // 2 + (buttonWidth % 2 > 0) # didn't want to import math.ceil just for that
        groupButtons.append(PushButton(app, text=sensor["address"],padx=padx, command=toggleGroup, args=[sensor["address"]], grid=[horizontalOffset,4,sensor["count"],1]))
        groupButtons[len(groupButtons)-1].bg = "#4d4d4d"
        groupButtons[len(groupButtons)-1].text_color = "white"
        groups[sensor["address"]]={"tracks":[i for i in range(horizontalOffset, horizontalOffset + sensor["count"])], "status":"normal"}
        horizontalOffset += sensor["count"] # shifted the next button to the right, covering it's tracks
        if horizontalOffset >= trackCount : break # all the active tracks have been covered

    if songTitle in pdPassthrough  : # if it ain't a multitrack song
        sendOSC("puredata","/"+songTitle) # alert pd that we are switching to synth
        playButton.hide() # we won't need this button
    else : playButton.show()
    if trackCount < 8 : # with less than 8 tracks, there isn't enough space on the top to fit theses two buttons
        sensorButton.tk.grid(row=6, column=0) # so we move them to the bottom
        calibButton.tk.grid(row=6, column=1)
    else :
        sensorButton.tk.grid(row=0, column=0) # else, they stay on top
        calibButton.tk.grid(row=0, column=1)
    app.width=83*trackCount # resizes the window to fit the tracks count (83px is the approximate width of a channel)
    statusText.value = ""

# Return scaled sensor volume (sensor/pd volume being 0~1000, UI volume 0~100)
def scaleVolume(sensorVolume, faderVolume) : return int(sensorVolume * faderVolume/100.)

# Check time spent since the last message was received, weed'em out if they are late
def purgeDisconnectedDevices(event=None) :
    currentTime = time()
    for OSCaddress in list(onlineDevices) : # using list instead of .keys() to avoid RuntimeError: dict changed size during iteration when using pop()
        if currentTime - onlineDevices[OSCaddress]["lastSeen"] > offlineThreeshold :
            # ~ sensorsWidgets[OSCaddress]["nightButton"].disable()
            # ~ sensorsWidgets[OSCaddress]["dayButton"].disable()
            sensorsWidgets[OSCaddress]["icon"].image = ICONS["sensorGray"]
            statusText.value = "L'appareil "+OSCaddress+" s'est déconnecté"
            sensorsWidgets[OSCaddress]["hostname"].value = ""
            sensorsWidgets[OSCaddress]["IP"].value = ""
            sendOSC(OSCaddress, "/sendID") # give it a chance to reconnect
        # ~ if currentTime - onlineDevices[OSCaddress]["lastSeen"] > offlineThreeshold * 10: # if still not there after a long time, cleanup
            # ~ onlineDevices.pop(OSCaddress)

# move the faders to their last received position
def updateUI(*event):
    # ~ if time()-startTime < 10 : return # wait 10s after launch to allow puredata to load tracks
    for trackNumber in range(0, len(sliders)) :
        if songs[songSelection.value]["tracks"][trackNumber]["status"] == "normal":
            sliders[trackNumber].value = str(songs[songSelection.value]["tracks"][trackNumber]["volume"]/10) # sliders are 0~100, OSC values 0~1000

def sendOSC(clientName, address, data=[]):
    msg = oscbuildparse.OSCMessage(address, None, data) # "typetags=None" automatically set the type of the bundled data based on their python type
    osc_send(msg, clientName)

# called on incoming OSC messages, dispatch them to puredata depending on expectedOSCmessages
def osc_incomingDataHandler (address, *data):
    # ~ print ("data received from {}: {}".format(address, data))
    trackNumber = 0 # will be incremented to map expectedOSCmessages over loaded tracks
    for message in expectedOSCmessages:
        if address == message["address"] : # if the message belongs to this definition
            for i in range(0, len(data)) :
                if trackNumber > len(songs[songSelection.value]["tracks"]) : 
                    statusText.value = "ERROR : not enough tracks on this song" # errors messages tends to be more intimidating in english
                    return
                songs[songSelection.value]["tracks"][trackNumber]["volume"]=data[i] # stores the received data for this track inside the songs variable
                if songSelection.value in pdPassthrough : # 
                    if data[i] > pdThreeshold and songs[songSelection.value]["tracks"][trackNumber]["status"] == "normal" :
                        sendOSC("puredata","/volume", [trackNumber, data[i]]) # only send data back to pd if above threeshold
                    else : sendOSC("puredata","/volume", [trackNumber, 0]) # set to zero if under threeshold
                elif songs[songSelection.value]["tracks"][trackNumber]["status"] == "normal":
                    if trackNumber in sampleTracks : playSample(trackNumber, data[i])
                    sendOSC("puredata","/volume", [trackNumber, data[i]])
                if address in onlineDevices : onlineDevices[address]["lastSeen"] = time() # if the device is already known to the server, store the time this message was received
                else : sendOSC("broadcast", "/sendID") # else, ask every device on the network to identify themselves since we don't have it's IP at this stage
                trackNumber += 1
            return
        else : trackNumber += message["count"] # not the droid we are looking for, skipping this group of tracks
        
# called on incoming "/myID" OSC messages
def osc_IDHandler (OSCaddress, hostname, IP, sensorCount):
    print("received ID : add ={}, hostname={}, IP={}, sensorCount={}".format(OSCaddress, hostname, IP, sensorCount))
    global onlineDevices
    if OSCaddress not in onlineDevices : # this device is still unknown
        osc_udp_client(IP, 8000, OSCaddress)
        onlineDevices[OSCaddress] = {"hostname" : hostname, "IP":IP, "sensorCount":sensorCount, "lastSeen":time()}
        statusText.value = "l'appareil "+OSCaddress+" est en ligne"
        sensorsWidgets[OSCaddress]["hostname"].value = hostname
        sensorsWidgets[OSCaddress]["IP"].value = IP
        sendOSC(OSCaddress, "/calibrateAmbiant") # auto calibrate ambiant on first connection to avoid spamming the network
    elif onlineDevices[OSCaddress]["hostname"] != hostname : statusText.value = "ERROR : two devices are competing for the same OSCaddress" # errors messages tends to be more intimidating in english
    else : statusText.value = "L'appareil "+OSCaddress+" s'est reconnecté"
    sensorsWidgets[OSCaddress]["nightButton"].enable()
    sensorsWidgets[OSCaddress]["dayButton"].enable()
    sensorsWidgets[OSCaddress]["icon"].image = ICONS["sensor"]


def exitCleanly(event=None):
    app.destroy()
    Popen("killall pd", shell=True)
    Popen("killall python3", shell=True) # the only time seppuku will not rinse the shame away
    raise SystemExit # will never get there
    
app.on_close(exitCleanly)
'''
--------------------------------------------------------------- ON START ----------------------------------------------------------
'''
Popen(pdCommand, shell=True) # start puredata
songs=updateSongList() # read files structure of the media folder 

# create the sensor window
sensorWindow = Window(app, layout="grid", bg="#4d4d4d", title="capteurs", width=140*len(expectedOSCmessages), height=280)
sensorWindow.repeat(offlineTimer, purgeDisconnectedDevices) # check if sensors are alive every offlineTimer seconds 
sensorsWidgets={}
for i in range(0, len(expectedOSCmessages)) :
    address = expectedOSCmessages[i]["address"]
    sensorsWidgets[address]={}
    sensorsWidgets[address]["icon"] = Picture(sensorWindow, image=ICONS["sensorGray"], grid=[i,1])
    sensorsWidgets[address]["label"] = Text(sensorWindow , color="white", text=expectedOSCmessages[i]["address"], grid=[i,2])
    sensorsWidgets[address]["dayButton"] = PushButton(sensorWindow, image=ICONS["day"], grid=[i,3], text="", enabled=False, command=calibrateSensor, args=[address, False])
    sensorsWidgets[address]["dayButton"] = PushButton(sensorWindow, image=ICONS["day"], grid=[i,3], text="", enabled=True, command=calibrateSensor, args=[address, False])
    sensorsWidgets[address]["nightButton"] = PushButton(sensorWindow, image=ICONS["night"], grid=[i,4], text="", command=calibrateSensor, args=[address, True], enabled=False)
    sensorsWidgets[address]["nightButton"] = PushButton(sensorWindow, image=ICONS["night"], grid=[i,4], text="", command=calibrateSensor, args=[address, True], enabled=True)
    sensorsWidgets[address]["hostname"] = Text(sensorWindow , color="white", text="", grid=[i,5])
    sensorsWidgets[address]["hostname"].text_size = 9
    sensorsWidgets[address]["IP"] = Text(sensorWindow , color="white", text="", grid=[i,6])

# Populate the transport box with play/pause prev and next buttons + the combo song selector
nextButton = PushButton(transportBox, image=ICONS["next"], text="", command=transport, args=["skipNext"], grid=[2,1])
playButton = PushButton(transportBox, image=ICONS["play"], text="", command=transport, args=["playPause"], grid=[1,1])
prevButton = PushButton(transportBox, image=ICONS["prev"], text="", command=transport, args=["skipPrev"], grid=[0,1])
songSelection = Combo(transportBox, options=songs.keys(), command=updateLayout, grid=[0,0,3,1])
songSelection.text_color = "white"; songSelection.text_size = 20
sensorButton = PushButton(app, image=ICONS["settings"], text="", command=sensorWindow.show, grid=[0,0])
calibButton = PushButton(app, image=ICONS["night"], text="", command=calibrateAllSensors, grid=[1,0])

# Define keyboard shortcuts :
app.tk.bind("<space>", lambda event : transport("playPause"))# space will toggle play/pause
app.tk.bind("<Control-space>", toggleAllTracks)# ctrl-space will toggle all tracks status between normal and manual
# ~ app.tk.bind("<Control-m>", lambda event: [setTrackStatus(t, "mute") for t in range(0, len(songs[songSelection.value]["tracks"]))])# ctrl-m will mute all tracks
app.tk.bind("<Control-m>", lambda event: [toggleGroup(g) for g in groups])# ctrl-m will mute all tracks
app.tk.bind("<Left>", lambda event : transport("skipPrev"))# Left arrow will load the previous song
app.tk.bind("<Right>", lambda event : transport("skipNext"))# Right arrow will load the next song
app.tk.bind("<Control-Return>", calibrateAllSensors)# Control-Enter will calibrate sensors
app.tk.bind("<Escape>", exitCleanly)# Control-Enter will calibrate sensors
# ~ [app.tk.bind("<F"+str(i+1)+">", lambda event : playSample(i+1)) for i in range(0, min(len(sampleTracks),12))]# Interestingly enough, this won't work because lambda assign the last (max) stored i value to each loop.
app.tk.bind("<F1>", lambda event : playSample(sampleTracks[0], 1000, True))
app.tk.bind("<F2>", lambda event : playSample(sampleTracks[1], 1000, True))
app.tk.bind("<F3>", lambda event : playSample(sampleTracks[2], 1000, True))
app.tk.bind("<F4>", lambda event : playSample(sampleTracks[3], 1000, True))
app.tk.bind("<F5>", lambda event : sendOSC("broadcast","/sendID")) #F5 refreshes sensors
sensorWindow.tk.bind("<Escape>", lambda event : sensorWindow.hide()) # ESC closes the sensor window

updateLayout(songSelection.value)

if __name__ == "__main__":
    # Start the threaded OSCserver, see OSCmessages.txt for details on OSC communication
    osc_startup()
    osc_udp_server(myIp, OSClisteningPort, "PiServer")
    osc_method("/myID", osc_IDHandler)
    # ~ osc_method("/start", osc_incomingDataHandler, argscheme=osm.OSCARG_ADDRESS + osm.OSCARG_DATAUNPACK) # currently not in use
    # ~ osc_method("/stop", osc_incomingDataHandler, argscheme=osm.OSCARG_ADDRESS + osm.OSCARG_DATAUNPACK)# currently not in use
    # ~ osc_method("/prev", osc_incomingDataHandler, argscheme=osm.OSCARG_ADDRESS + osm.OSCARG_DATAUNPACK)# currently not in use
    # ~ osc_method("/next", osc_incomingDataHandler, argscheme=osm.OSCARG_ADDRESS + osm.OSCARG_DATAUNPACK)# currently not in use
    for message in expectedOSCmessages : osc_method(message["address"], osc_incomingDataHandler, argscheme=osm.OSCARG_ADDRESS + osm.OSCARG_DATAUNPACK)
    print("Listening on "+myIp+":"+str(OSClisteningPort))
    sendOSC("broadcast","/sendID") # request ID from all connected devices
    sendOSC("broadcast","/ledON") # turn the talent led on for every sensor by default
    app.repeat(updateUIdelay, updateUI) # periodically update slider values
    app.display() # start Tk
